import typing as t
from contextlib import contextmanager
from sqlalchemy.exc import MultipleResultsFound, NoResultFound

from aerie.exceptions import NoResultsError, TooManyResultsError


@contextmanager
def convert_exceptions() -> t.Generator[None, None, None]:
    try:
        yield
    except MultipleResultsFound as exc:
        raise TooManyResultsError from exc
    except NoResultFound as exc:
        raise NoResultsError from exc



from sqlalchemy.sql import select

from .database import Aerie
from .exceptions import AerieError, NoResultsError, TooManyResultsError
from .models import Model, metadata
from .session import DbSession, Page

__all__ = [
    'Aerie',
    'DbSession',
    'Page',
    'Model',
    'metadata',
    'TooManyResultsError',
    'NoResultsError',
    'AerieError',
    'select',
]



from __future__ import annotations

import math
import typing as t
from sqlalchemy import exists, select
from sqlalchemy.exc import MultipleResultsFound
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.sql import Select
from sqlalchemy.sql.functions import func

from aerie.exceptions import TooManyResultsError
from aerie.utils import convert_exceptions

M = t.TypeVar('M')


class Page(t.Generic[M]):
    def __init__(self, rows: t.Sequence[M], total_rows: int, page: int, page_size: int) -> None:
        self.rows = rows
        self.total_rows = total_rows
        self.page = page
        self.page_size = page_size
        self._pointer = 0

    @property
    def total_pages(self) -> int:
        """Total pages in the row set."""
        return math.ceil(self.total_rows / self.page_size)

    @property
    def has_next(self) -> bool:
        """Test if the next page is available."""
        return self.page < self.total_pages

    @property
    def has_previous(self) -> bool:
        """Test if the previous page is available."""
        return self.page > 1

    @property
    def has_other(self) -> bool:
        """Test if page has next or previous pages."""
        return self.has_next or self.has_previous

    @property
    def next_page(self) -> int:
        """Next page number. Always returns an integer.
        If there is no more pages the current page number returned."""
        return min(self.total_pages, self.page + 1)

    @property
    def previous_page(self) -> int:
        """Previous page number. Always returns an integer.
        If there is no previous page, the number 1 returned."""
        return max(1, self.page - 1)

    @property
    def start_index(self) -> int:
        """The 1-based index of the first item on this page."""
        if self.page == 1:
            return 1
        return (self.page - 1) * self.page_size + 1

    @property
    def end_index(self) -> int:
        """The 1-based index of the last item on this page."""
        return min(self.start_index + self.page_size - 1, self.total_rows)

    def __iter__(self) -> t.Iterator[M]:
        return iter(self.rows)

    def __next__(self) -> M:
        if self._pointer == len(self.rows):
            raise StopIteration
        self._pointer += 1
        return self.rows[self._pointer - 1]

    def __getitem__(self, item: int) -> M:
        return self.rows[item]

    def __len__(self) -> int:
        return len(self.rows)

    def __bool__(self) -> bool:
        """A shortcut to check if page has more than one pages.
        Useful in templates to check if the pagination should be rendered or not."""
        return self.total_pages > 1

    def __str__(self) -> str:
        return (
            f'Page {self.page} of {self.total_pages}, rows {self.start_index} - {self.end_index} of {self.total_rows}.'
        )

    def __repr__(self) -> str:
        return f'<Page: page={self.page}, total_pages={self.total_pages}>'


class ExecutableQuery:
    def __init__(self, session: DbSession, stmt: Select, params: t.Mapping = None) -> None:
        self._session = session
        self._stmt = stmt
        self._params = params

    async def all(self) -> t.Sequence:
        result = await self._session.execute(self._stmt, self._params)
        return result.scalars().all()

    async def first(self) -> t.Optional[M]:
        result = await self._session.execute(self._stmt, self._params)
        return result.scalars().first()

    async def one(self) -> M:
        """Get exactly one result or raise.

        :raise MultipleResultsFound - when more that one row matched
        :raise NoResultFound - when no rows matched
        """

        with convert_exceptions():
            result = await self._session.execute(self._stmt, self._params)
            return result.scalars().one()

    async def one_or_none(self) -> t.Optional[M]:
        """Get one result or return None if none found..

        :raise MultipleResultsFound"""

        try:
            result = await self._session.execute(self._stmt, self._params)
            return result.scalars().one_or_none()
        except MultipleResultsFound as exc:
            raise TooManyResultsError() from exc

    async def count(self) -> int:
        stmt = select(func.count('*')).select_from(self._stmt)
        return await self._session.scalar(stmt, self._params)

    async def exists(self) -> bool:
        return await self._session.scalar(select(exists(self._stmt)), self._params)

    async def paginate(self, page: int = 1, page_size: int = 50) -> Page:
        total = await self._session.query(self._stmt, self._params).count()
        offset = (page - 1) * page_size

        stmt = self._stmt.limit(page_size).offset(offset)
        rows = await self._session.query(stmt, self._params).all()
        return Page(rows, total, page, page_size)


class DbSession(AsyncSession):
    def select(self, model: t.Type[M]) -> Select:
        return select(model)

    def query(self, stmt: Select, params: t.Mapping = None) -> ExecutableQuery:
        return ExecutableQuery(self, stmt, params)



import typing as t
from sqlalchemy import exists, func, select
from sqlalchemy.engine import Result, Row
from sqlalchemy.ext.asyncio import AsyncEngine
from sqlalchemy.sql import Executable

from aerie.utils import convert_exceptions

R = t.TypeVar('R')


class ExecutableQuery:
    def __init__(self, engine: AsyncEngine, stmt: Executable, params: t.Mapping = None) -> None:
        self._engine = engine
        self._stmt = stmt
        self._params = params

    async def first(self) -> t.Optional[Row]:
        async with self._engine.begin() as connection:
            result = await connection.execute(self._stmt, self._params)
            return result.first()

    async def all(self) -> t.List[Row]:
        async with self._engine.begin() as connection:
            result = await connection.execute(self._stmt, self._params)
            return result.all()

    async def one(self) -> Row:
        with convert_exceptions():
            async with self._engine.begin() as connection:
                result = await connection.execute(self._stmt, self._params)
                return result.one()

    async def one_or_none(self) -> t.Optional[Row]:
        with convert_exceptions():
            async with self._engine.begin() as connection:
                result = await connection.execute(self._stmt, self._params)
                return result.one_or_none()

    async def scalars(self, index: int = 0) -> t.List[t.Any]:
        async with self._engine.begin() as connection:
            result = await connection.execute(self._stmt, self._params)
            return result.scalars(index=index).all()

    async def scalar(self) -> t.Any:
        with convert_exceptions():
            async with self._engine.begin() as connection:
                result = await connection.execute(self._stmt, self._params)
                return result.scalar_one()

    async def scalar_or_none(self) -> t.Optional[t.Any]:
        with convert_exceptions():
            async with self._engine.begin() as connection:
                result = await connection.execute(self._stmt, self._params)
                return result.scalar_one_or_none()

    async def execute(self) -> Result:
        async with self._engine.begin() as connection:
            return await connection.execute(self._stmt, self._params)

    async def count(self) -> int:
        stmt = select(func.count('*')).select_from(self._stmt)
        async with self._engine.begin() as connection:
            result = await connection.execute(stmt, self._params)
            count = result.scalar()
        return int(count) if count else 0

    async def exists(self) -> bool:
        async with self._engine.begin() as connection:
            result = await connection.execute(select(exists(self._stmt)), self._params)
            return result.scalar() is True



import sqlalchemy as sa
import typing as t
from sqlalchemy import MetaData
from sqlalchemy.ext.asyncio import AsyncEngine, create_async_engine
from sqlalchemy.orm import sessionmaker
from sqlalchemy.sql import Executable

from aerie.models import metadata as shared_metadata
from aerie.query import ExecutableQuery
from aerie.session import DbSession

_IsolationLevel = t.Literal['SERIALIZABLE', 'REPEATABLE READ', 'READ COMMITTED', 'READ UNCOMMITTED', 'AUTOCOMMIT']


class Aerie:
    def __init__(
        self,
        url: str,
        echo: bool = False,
        isolation_level: _IsolationLevel = None,
        json_serializer: t.Callable = None,
        json_deserializer: t.Callable = None,
        pool_size: int = 5,
        pool_timeout: int = 30,
        max_overflow: int = 10,
        metadata: MetaData = None,
    ) -> None:
        self.url = url
        self.metadata = metadata or shared_metadata
        self._engine = create_async_engine(
            url,
            echo=echo,
            isolation_level=isolation_level,
            json_serializer=json_serializer,
            json_deserializer=json_deserializer,
        )
        self._session_maker = sessionmaker(bind=self._engine, class_=DbSession, expire_on_commit=False)

    def session(self) -> t.AsyncContextManager[DbSession]:
        """Create a new session object."""
        return self._session_maker()

    def transaction(self) -> AsyncEngine._trans_ctx:
        """Establish a new transaction."""
        return self._engine.begin()

    def query(self, stmt: Executable, params: t.Mapping = None) -> ExecutableQuery:
        return ExecutableQuery(self._engine, stmt, params)

    async def create_tables(self, tables: t.List[t.Union[str, sa.Table]] = None) -> None:
        if tables:
            tables = [self.metadata.tables[table] if isinstance(table, str) else table for table in tables]
        async with self._engine.begin() as connection:
            await connection.run_sync(self.metadata.create_all, tables=tables)

    async def drop_tables(self, tables: t.List[t.Union[str, sa.Table]] = None) -> None:
        if tables:
            tables = [self.metadata.tables[table] if isinstance(table, str) else table for table in tables]

        async with self._engine.begin() as connection:
            await connection.run_sync(self.metadata.drop_all, tables=tables)
